

#ifdef WIN32
#include <winsock2.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <time.h>

void taskDelay(int);

int GANILtransferTxData (char *, unsigned long, int);
int GANILtransferTxRestart ();
int GANILtransferSetVerbose(int);
int GANILtransferNice(int);
int GANILtransferUseOverlap(int);

int transferTxData (char *, unsigned long, int);
int transferTxRestart ();
int transferTxReset ();
int transferBlockSize(int);
int transferMode(int);
int transferPort(int);
int transferInit (char *);
void transferClose ();
int transferSetVerbose(int);
int transferSetUser(int);
int transferNice(int);
int transferUseOverlap(int);
int transferTxWait();


#ifdef SOLARIS
#include <unistd.h>
#endif

#include <signal.h>

void signal_block ()
{
   sigset_t signalSet;

//    mask off signals handled by the main thread   - taken from general DA source

   sigemptyset(&signalSet);
   sigaddset(&signalSet, SIGINT);
   sigaddset(&signalSet, SIGUSR1);
   sigaddset(&signalSet, SIGUSR2);
   sigprocmask(SIG_BLOCK, &signalSet, NULL);
}

void Usage(char *progname) {
    fprintf(stderr,"Usage\n%s -n [server] -p [port] -m [mode] -t [ticks] -s [buffersize] -b [num]\n",progname);
    fprintf(stderr,"\tserver is the IP address or name of server\n");
    fprintf(stderr,"\tport is the TCP port to use\n");
    fprintf(stderr,"\tmode is the protocol transfer option\n");
    fprintf(stderr,"\tticks is the delay between sending data blocks in millisecs\n");
    fprintf(stderr,"\ts is data block size in Kbytes\n");
    fprintf(stderr,"\tb is the number of blocks sent at each tick\n");
    fprintf(stderr,"Defaults are localhost; 10305; 3; 10; 16K and 1\n");
    fprintf(stderr,"mode = 5 is for GANIL\n");
    exit(1);
}
        
int main(int argc, char **argv) {

    int Buffer[64*1024];
    char *TS_Server= (char*)"localhost";
    int Port=10305;
    unsigned long stream = 1;
    int Mode=3;
    int DelayTicks=10;     /* 10 millisecs  */
    int NumBlocks=1;
    int s;
    int i;
    int BufferSize=16*1024;
    int num;
    int b;

    struct timeval Time;
    int InitialTime;
    int diff;

#ifndef WIN32
    struct timespec time_request;
#endif

    if (argc >1) {
        for(i=1;i <argc;i++) {
            if ( (argv[i][0] == '-') || (argv[i][0] == '/') ) {
                switch(tolower(argv[i][1])) {
                    case 'n':
                        TS_Server = argv[++i];
                        break;
                    case 'p':
                        Port = atoi(argv[++i]);
                        break;
                    case 'm':
                        Mode = atoi(argv[++i]);
                        break;
                    case 't':
                        DelayTicks = atoi(argv[++i]);
                        break;
                    case 'b':
                        NumBlocks = atoi(argv[++i]);
                        break;
                    case 's':
                        BufferSize = atoi(argv[++i])*1024;
                        break;
                    default:
                        Usage(argv[0]);
                        break;
                }
            }
            else
                Usage(argv[0]);
        }
    }

  if (Mode == 5) {
    (void) GANILtransferSetVerbose(1);
    (void) GANILtransferUseOverlap(1);
    (void) GANILtransferNice(1);
    (void) GANILtransferTxRestart();
    printf ("Resetting communication with GANIL server\n");

  } else {
 //   (void) transferSetVerbose(1);
    (void) transferUseOverlap(1);
    (void) transferNice(1);
    (void) transferBlockSize(BufferSize);
    (void) transferMode(Mode);
    (void) transferPort(Port);
    (void) transferInit(TS_Server);
    printf ("Resetting communication with MIDAS server\n");
    (void) transferTxReset();
  }

    
#ifdef WIN32
    time(&Time.tv_sec);
#else
    gettimeofday(&Time, NULL);
#endif

    for (i=0; i<(BufferSize/4); i++) {
#ifdef WIN32
        Buffer[i] = rand();
#else
        Buffer[i] = random();
#endif
    }

    InitialTime = Time.tv_sec;
    printf("Start at %d\n", InitialTime);

    num = 0;
    
    for (;;)
    {

       if (DelayTicks) {
#ifdef WIN32
           Sleep((DWORD)DelayTicks);
#endif
#ifdef SOLARIS
           (void) usleep(DelayTicks * 1000);
#endif
#ifdef LINUX
           time_request.tv_sec = DelayTicks/1000;
           time_request.tv_nsec = (DelayTicks - (time_request.tv_sec * 1000)) * 1000000;
           i = nanosleep(&time_request, NULL);
#endif
#ifdef LYNXOS
           time_request.tv_sec = DelayTicks/1000;
           time_request.tv_nsec = (DelayTicks - (time_request.tv_sec * 1000)) * 1000000;
           i = nanosleep(&time_request, NULL);
#endif
        }
         
     for (b=0; b<NumBlocks; b++)
     {
       if (Mode == 5) {
         s = GANILtransferTxData ((char *)&Buffer[8] , stream, BufferSize-32);
       } else {
         s = transferTxData ((char *)&Buffer[8] , stream, BufferSize-32);
       }
         num++;
         if ((num/1000)*1000 == num) {

#ifdef WIN32
             time(&Time.tv_sec);
#else
             gettimeofday(&Time, NULL);
#endif

             diff = Time.tv_sec-InitialTime;
             printf ("sent %d blocks after %d\n", num, diff);
         }
      }

    }

}
