/*******************************************************************************
*  transfer.c            Interface procedures for general DA program use
*/


#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#endif

#if (defined UNIX || defined GWIN32)
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#endif

#ifdef LYNXOS
#include <common.h>
#include <socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <resource.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <errno.h>
#include "msg.h"

#include "transfer.h"

#ifdef USEPTHREAD
#define VERSION "3.8T"
#define _POSIX_C_SOURCE 199506L
#else
#define VERSION "3.8"
#endif

#define MAXUSERS 4
int CurrentUser = 0;

unsigned long transferSequence[MAX_STREAM+1];    /*  stream block sequence number */
unsigned long UsertransferSequence[MAXUSERS][MAX_STREAM+1];


/* Default port number is based on MIDAS_TAPE_SERVER (RPC program number) + 100
 * i.e. if MIDAS_TAPE_SERVER is 28000205 TS_TCP_PORT is 10305
*/

#define TS_TCP_PORT 10305

/* for Mode = 4  (GSI/MBS compatible) the default port is 6500 */

#define MBS_TCP_PORT 6500

/* for Mode = 6 (LNL/XDAQ compatible) the default port is 50001 */

#define XDAQ_TCP_PORT 50001

unsigned short Port = 0;       /* not defined initially */
unsigned short UserPort[MAXUSERS] = {0, 0, 0, 0};

char tsHost[80];
unsigned char UsertsHost[MAXUSERS][80];

char MsgBuf[256];

char Ack [sizeof(ACK)];
char Restart [1024];           // size for restart message

int verbose = 0;

int BlockSize = 16384;         //  16K bytes is default
int UserBlockSize[MAXUSERS] = {16384, 16384, 16384, 16384};

int XferMode = 1;              //  default - 32 byte protocol + dacq
int UserXferMode[MAXUSERS] = {1, 1, 1, 1};

#define MINSNDBUFSIZE 32   /*  units Kbytes  */
#define MAXSNDBUFSIZE 4*1024

#ifdef LYNXOS
#include <signal.h>
struct sigaction sa;
int sigpipe=0;          // set when signal handler for SIGPIPE installed
#endif

#ifdef WIN32
    WSADATA wsaData;
#else
    #define SOCKET int
    #define SOCKET_ERROR -1
#endif

#ifdef USEPTHREAD
#include <pthread.h>
#include <sched.h>

pthread_mutex_t M1[MAXUSERS];
pthread_cond_t  C1[MAXUSERS];
pthread_mutex_t M2[MAXUSERS];
pthread_cond_t  C2[MAXUSERS];

int pthreads_started = 0;

int pthread_running[MAXUSERS];

//    arguments for request passed to thread
SOCKET pthread_tsSocket[MAXUSERS];
char * pthread_blockptr[MAXUSERS];
int pthread_len[MAXUSERS];

void signal_block();
#endif

//    these are only of use when threads enabled
int OverlapMode = 0;
int NiceValue = 0;

struct sockaddr_in  server;

SOCKET tsSocket = 0;
SOCKET UsertsSocket[MAXUSERS] = {0, 0, 0, 0};

#ifdef UNIX

int msgDefineMessage (u_int id, u_int xclass, u_int level, char *source, char *body) {

 // printf("id=%d, class=%d, level=%d, source=%s: %s\n",id, xclass, level, source, body);
  return OK;
}
#endif

#ifdef WIN32
BOOL WINAPI DllMain (HANDLE hModule, DWORD dwFunction, LPVOID lpNot) {

    return TRUE;
}
#endif

#ifdef WIN32
void ReportError () {

  LPVOID lpMsgBuf;
  DWORD ErrNo;
  
       ErrNo = WSAGetLastError();
       FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      ErrNo,
                      0,
                      (LPTSTR) &lpMsgBuf,
                      0,
                      NULL
                      );
                      
       sprintf(MsgBuf, "Transfer Error %d - %s", ErrNo, lpMsgBuf);
       fprintf(stderr, "%s", MsgBuf);
       msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
       LocalFree (lpMsgBuf);
}
#else
void ReportError () {
       perror("Transfer Error - ");
       sprintf(MsgBuf, "Error %d", errno);
       msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
}
#endif

#ifdef USEPTHREAD
/*******************************************************************************
    sends any pending buffer
    loops until all data has been sent  (or ERROR)

    there is one thread for each USER
*/
void * Thread_Send(void * i)
{
   SOCKET S;
   char * p;
   int l;

   int rc, done, left;
   int num = 0;
   int nicevalue = 0;
   char *ErrMsg;

#ifdef LYNXOS
   int id, prty1, prty2;
#endif

   num = (int) i;   // this is the thread (USER) number

   if (verbose) printf("pthread %d running\n", num);

   signal_block();   // block signals handled by the main thread

   for (;;) {

      if (verbose) printf("Thread_Send waiting\n");

//    wait on the conditional variable C1 for work to be done

      rc = pthread_mutex_lock(&M1[num]);
      if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_lock 1: %s\n", ErrMsg);}

      while (!pthread_running[num]) {
          rc = pthread_cond_wait (&C1[num], &M1[num]);   // this will block until signaled
          if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_cond_wait 1: %s\n", ErrMsg); sleep(1);}
      }

      rc = pthread_mutex_unlock(&M1[num]);
      if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_unlock 1: %s\n", ErrMsg);}

      if (verbose) printf("Thread_Send running\n");

      if (NiceValue != nicevalue) {    // check if nice has changed 
          nicevalue = NiceValue;  
#ifndef LYNXOS
          nice(NiceValue);
#else
          id = BUILDPID(getpid(), getstid());
          prty1 = getpriority(PRIO_PROCESS, id);
          (void) setpriority(PRIO_PROCESS, id, prty1 - NiceValue);
          prty2 = getpriority(PRIO_PROCESS, id);
          printf("Thread %d priority changed from %d to %d\n", num, prty1, prty2);
#endif
      }

/*    pick up request argunments  */

      S = pthread_tsSocket[num];
      p = pthread_blockptr[num];
      l = pthread_len[num];

      done = 0;
      left =  l;

      while (done < l) {
         rc = send(S, p+done, left, 0);
         if (rc == SOCKET_ERROR) {
            ReportError();
            sprintf(MsgBuf, "send() failed: ");
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            transferClose();
            break;
         }
         done += rc;
         left -= rc;
       }

       if (verbose) printf("Thread_Send finishing\n");

//    signal the conditional variable C2 that the work has been completed

       rc = pthread_mutex_lock(&M2[num]);
       if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_lock 2: %s\n", ErrMsg);}

       pthread_running[num] =0;          // clear the running flag

       rc = pthread_cond_signal(&C2[num]);   // signal the thread
       if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_cond_signal 2: %s\n", ErrMsg);}

       rc = pthread_mutex_unlock(&M2[num]);  //   allow it to run
       if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_unlock 2: %s\n", ErrMsg);}
    }

    return ((void *)0);
}
/*******************************************************************************/
void do_send_thread(SOCKET S, char * p, int l)
{
    int rc;
   char *ErrMsg;

//    start the send thread
//    signal the conditional variable C1 that there is work to be done

        rc = pthread_mutex_lock(&M1[CurrentUser]);
        if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_lock 3: %s\n", ErrMsg);}

        pthread_tsSocket[CurrentUser] = S;   /*  save the thread arguments  */
        pthread_blockptr[CurrentUser] = p;
        pthread_len[CurrentUser] = l;

        pthread_running[CurrentUser] =1;         //  set the running flag

        rc = pthread_cond_signal(&C1[CurrentUser]);   // signal the thread
        if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_cond_signal 3: %s\n", ErrMsg);}

        rc = pthread_mutex_unlock(&M1[CurrentUser]);  // allow it to run
        if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_unlock 3: %s\n", ErrMsg);}
}
/********************************************************************************
*    transferTxWait       wait for any pending transfer to complete
*
*/
int transferTxWait ()
{
    int rc;
    int retval;
   char *ErrMsg;

        if (verbose) printf("TransferTxWait waiting\n");

//     if (!pthread_running[CurrentUser]) return 0;

//    wait on the conditional variable C2 for completion

       retval = 0;

        rc = pthread_mutex_lock(&M2[CurrentUser]);
        if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_lock 4: %s\n", ErrMsg);}

        while (pthread_running[CurrentUser]) {
            rc = pthread_cond_wait(&C2[CurrentUser], &M2[CurrentUser]);   // this will block until signaled
            if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_cond_wait 4: %s\n", ErrMsg); sleep(1);}
            retval = 1;
        }

        rc = pthread_mutex_unlock(&M2[CurrentUser]);
        if (rc != 0) {ErrMsg = strerror(rc); printf ("error in pthread_mutex_unlock 4: %s\n", ErrMsg);}

        if (verbose) printf("TransferTxWait complete\n");

     return retval;
}
#else
int transferTxWait () {return 0;}              //   dummy for the non thread case
#endif
/*******************************************************************************
    same parameter list as send
    loops until all data has been sent  (or ERROR)
*/
int do_send(SOCKET S, char * p, int l)
{
   int rc, done, left;

    done = 0;
    left =  l;

    while (done < l) {
        rc = send(S, p+done, left, 0);
        if (rc == SOCKET_ERROR)
             return rc;
        done += rc;
        left -= rc;
    }

    return l;
}

/*******************************************************************************
    same parameter list as recv
    loops until all data has been received  (or ERROR)
*/
int do_recv(SOCKET S, char * p, int l, int t)
{
   int rc, done, left;

    done = 0;
    left =  l;
    while (done < l) {
        rc = recv(S, p+done, left, t);
        if (rc == SOCKET_ERROR)
             return rc;
        if (rc == 0) break;
        done += rc;
        left -= rc;
    }

    return l;
}

/*******************************************************************************
* transferInit   initialise interface
*            return OK or ERROR
*    tsIPStr is a char* to the name of the Tape Server
*/
int transferInit (char *tsIPStr)
{
    int rs;
    
    strcpy (tsHost, tsIPStr);

    rs = transferTxRestart();

    return rs;
}
/*******************************************************************************
* transferBlockSize        allows a custom size to be set
*/
int transferBlockSize(int n)
{
    BlockSize = n;
    printf("Setting Transfer Block Size %d\n",BlockSize);
    return OK;
}
/*******************************************************************************
* transferMode        allows the mode to be set
*/
int transferMode(int n)
{
    XferMode = n;
    printf("Setting Transfer Mode %d\n",XferMode);
    return OK;
}
/*******************************************************************************
* transferUseOverlap        enables the overlap mode - needs threads
*/
int transferUseOverlap(int n)
{
    OverlapMode = n;
    printf("Setting Overlap Mode %d\n",OverlapMode);
    return OK;
}
/*******************************************************************************
* transferNice        sets the nice value - needs threads
*/
int transferNice(int n)
{
    NiceValue = n;
    printf("Setting nice %d\n",NiceValue);
    return OK;
}
/*******************************************************************************
* transferPort        allows a custom port to be set
*/
int transferPort(int n)
{
    Port = (unsigned short) n;
    return OK;
}
/*******************************************************************************
* transferSetVerbose        allows more debug information 
*/
int transferSetVerbose(int n)
{
    verbose = (unsigned short) n;
    return OK;
}
/*******************************************************************************
* transferSetUser        allows current user to be set 
*/
int transferSetUser(int n)
{
    int i;

    if (n>MAXUSERS) return ERROR;
    if (n == CurrentUser) return OK;

/*    change user  */

/*    save current user  */
    UserPort[CurrentUser] = Port;
    UserBlockSize[CurrentUser] = BlockSize;
    UserXferMode[CurrentUser] = XferMode;
    UsertsSocket[CurrentUser] = tsSocket;

    for (i = 0; i <= MAX_STREAM; i++)
    {
        UsertransferSequence[CurrentUser][i] = transferSequence[i];
    }

    for (i = 0; i < 80; i++)
    {
        UsertsHost[CurrentUser][i] = tsHost[i];
    }

/*    load new user  */
    CurrentUser = (unsigned short) n;
    Port = UserPort[CurrentUser];
    BlockSize = UserBlockSize[CurrentUser];
    XferMode = UserXferMode[CurrentUser];
    tsSocket = UsertsSocket[CurrentUser];

    for (i = 0; i <= MAX_STREAM; i++)
    {
        transferSequence[i] = UsertransferSequence[CurrentUser][i];
    }

    for (i = 0; i < 80; i++)
    {
        tsHost[i] = UsertsHost[CurrentUser][i];
    }

    return OK;
}
/*******************************************************************************
* tsOpenClient            initial function to establish host/client connection
*                        create a socket
*                        find the host
*                        create a client
*                        returns client pointer or NULL
*/
int tsOpenClient (char *tsIPStr)
{         
       int retval;
       struct hostent *hp;
       unsigned int addr;
       struct sockaddr_in server;
       int sndsize, sndreq, temp, length;
#ifdef USEPTHREAD
       pthread_t thread;
       pthread_attr_t attr;
       int i;
#endif

    sprintf(MsgBuf, "TCP transfer library version %s",VERSION);
    printf ("%s\n", MsgBuf);
    msgDefineMessage (0, MSG_C_EGLK, MSG_S_INFORMATION, NULL, MsgBuf);

#ifdef USEPTHREAD

/*    start threads which will handle the data block send     */

    if (!pthreads_started) {

      printf("Threads enabled\n");

      retval = pthread_attr_init(&attr);
      if (retval != 0) perror("thread attribute");
      retval = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);  // we never use join
      if (retval != 0) perror("thread attribute");

      for (i = 0; i < MAXUSERS; i++)
      {
         pthread_running[i] = 0;

         pthread_mutex_init(&M1[i], NULL);
         pthread_mutex_init(&M2[i], NULL);
         pthread_cond_init(&C1[i], NULL);
         pthread_cond_init(&C2[i], NULL);

         retval = pthread_create(&thread, &attr, Thread_Send, (void *)i);
         if (retval != 0) {
            ReportError();
            sprintf(MsgBuf, "Client: Error creating pthread: ");
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            return ERROR;
         }
      }

      pthread_attr_destroy(&attr);
      pthreads_started =1;

      sleep(2);     // allows the threads to get running 
   }
#endif

#ifdef LYNXOS
    if (!sigpipe) {
        sa.sa_handler = SIG_IGN;
        sigemptyset(&sa.sa_mask);
        sa.sa_flags=0;

        (void) sigaction(SIGPIPE, &sa, NULL);
        sigpipe=1;
    }
#endif

#ifdef WIN32
    if ((retval = WSAStartup(MAKEWORD( 2, 2),&wsaData)) != 0) {
        WSASetLastError(retval);
        ReportError();
        sprintf(MsgBuf, "WSAStartup failed:");
        fprintf (stderr, "%s\n", MsgBuf);
        msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
        transferClose();
        return ERROR;
    }
#endif

     if (Port == 0) {
         Port = TS_TCP_PORT;
         if (XferMode == 4) Port = MBS_TCP_PORT;
         if (XferMode == 6) Port = XDAQ_TCP_PORT;
     }

     memset(&server,0,sizeof(server));
     server.sin_port = htons(Port);
    
    // Attempt to detect if we should call gethostbyname() or gethostbyaddr()

    if (isalpha(tsIPStr[0])) {   /* server address is a name */
        hp = gethostbyname(tsIPStr);
        if (hp == NULL ) {
            sprintf(MsgBuf,"Cannot resolve address [%s]: ", tsIPStr);
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            transferClose();
            return ERROR;
        }
        server.sin_family = hp->h_addrtype;
        memcpy(&(server.sin_addr),hp->h_addr,hp->h_length);
    }
    else  { /* Convert nnn.nnn address to a usable one */
        addr = inet_addr(tsIPStr);
        server.sin_family = AF_INET;
        server.sin_addr.s_addr = addr;
    }

    tsSocket = socket(AF_INET, SOCK_STREAM, 0); /* Open a TCP socket */
    if (tsSocket < 0 ) {
        ReportError();
        sprintf(MsgBuf, "Client: Error Opening socket: ");
        fprintf (stderr, "%s\n", MsgBuf);
        msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
        transferClose();
        return ERROR;
    }
    

    length=sizeof(sndsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stderr, "%s\n", MsgBuf);
        msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
    }
    temp = sndsize;

#ifdef LYNXOS

    sprintf(MsgBuf, "TCP socket send buffer is %d bytes",sndsize);
    fprintf (stdout, "%s\n", MsgBuf);
    msgDefineMessage (0, MSG_C_EGLK, MSG_S_INFORMATION, NULL, MsgBuf);

#else

    for (sndreq = MINSNDBUFSIZE; sndreq <= MAXSNDBUFSIZE; sndreq = sndreq+MINSNDBUFSIZE) {
        sndsize = sndreq*1024;
        if (setsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,sizeof(sndsize)) == SOCKET_ERROR) {
//            ReportError();
//            sprintf(MsgBuf, "setsockopt() failed:  ");
//            fprintf (stderr, "%s\n", MsgBuf);
//            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
              break;
        }
    }

    length=sizeof(sndsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stderr, "%s\n", MsgBuf);
        msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
    }
    
    sprintf(MsgBuf, "TCP socket send buffer was %d - now %d",temp,sndsize);
    fprintf (stdout, "%s\n", MsgBuf);
    msgDefineMessage (0, MSG_C_EGLK, MSG_S_INFORMATION, NULL, MsgBuf);


    sprintf(MsgBuf, "TCP socket created OK - now connecting to %s",tsIPStr);
    fprintf (stdout, "%s\n", MsgBuf);
    msgDefineMessage (0, MSG_C_EGLK, MSG_S_INFORMATION, NULL, MsgBuf);

#endif

    if (connect(tsSocket,(struct sockaddr*)&server,sizeof(server)) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "connect() failed:  ");
        fprintf (stderr, "%s\n", MsgBuf);
        msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
        transferClose();
        return ERROR;
    }

/*
    if (fcntl(tsSocket, F_SETFL, O_NONBLOCK) == -1) {
        ReportError();
        sprintf(MsgBuf, "fcntl() failed:  ");
        fprintf (stderr, "%s\n", MsgBuf);
    }
*/

    sprintf(MsgBuf, "Connected to %s port %d",tsIPStr, server.sin_port);
    fprintf (stdout, "%s\n", MsgBuf);
    msgDefineMessage (0, MSG_C_EGLK, MSG_S_INFORMATION, NULL, MsgBuf);

    return OK;
}
/*******************************************************************************
* transferClose            final function to close host/client connection
*                        destroys a socket
*/
void transferClose ()
{

    if (tsSocket != 0) {
#ifdef WIN32
             closesocket (tsSocket);
#else
             close (tsSocket);
#endif
             tsSocket = 0;
    }
    
#ifdef WIN32
        WSACleanup();
#endif
}
/*******************************************************************************
* transferStatus          returns state of socket  = 0 for no connection
*           
*/
int transferStatus ()
{
    return (tsSocket);
}
/*******************************************************************************
* clientInit   initialise 
*              initialise control structures
*              initialise client structure
*/
int clientInit (char *host)
{
    int i;
        int rs;


    for (i = 1; i <= MAX_STREAM; i++)
    {
        transferSequence[i] = 0;
        if (XferMode == 6) transferSequence[i] = 1;
    }

    transferClose ();
    strcpy (tsHost, host);
    rs = tsOpenClient (tsHost);
    return (rs);
}
/*******************************************************************************
* transferTxRestart   transmit the line restart message
*/
int transferTxRestart ()
{
    int rs;

     rs = clientInit (tsHost);
     if (rs != OK) return rs;

     rs = transferTxData (0, 0, -1);
     return rs;
}
/*******************************************************************************
* transferTxReset     reset all 4 streams
*                 transmit the line restart datagram
*/
int transferTxReset ()
{
    int i;

    for (i = 1; i <= MAX_STREAM; i++)
    {
        transferSequence[i] = 0;
        if (XferMode == 6) transferSequence[i] = 1;
    }

    return OK;
}
/*******************************************************************************
*/
int transferState ()
{
#ifdef USEPTHREAD
    return pthread_running[CurrentUser];
#else
    return 0;
#endif
}
/********************************************************************************
* transferTxData       Transmit the data 
*                  Check if stream is on
*                  Check window
*/
int transferTxData (char *data, unsigned long stream, int length)
{
        int retval, len;
        HEADER * blockptr;
        MBS_HEADER * mbsblockptr;
        I2O_SEND_TO_BU_MESSAGE_FRAME * xdaqblockptr;
        ACK * ackptr;
        unsigned short flags;
        int i;
        unsigned char * ptr;
        unsigned char b1, b2, b3, b4;

/*    if no connection currently exists - open one */


    if (tsSocket == 0)  (void)transferTxRestart();
    if (tsSocket == 0)  return ERROR;

        if (stream > MAX_STREAM) return -3;     /* check valid stream number */

/*    send the data block  */

      if (length < 0) {

/*    length < 0 is used for the RESTART message  */

          if (XferMode == 6) return OK;  /*  no RESTART message is used by Mode = 6 (XDAQ/LNL)  */

          blockptr = (HEADER*) Restart;

          if (XferMode == 4) {                /*  GSI/MBS mode  */
              mbsblockptr = (MBS_HEADER*) Restart;
              mbsblockptr->mbs_endian = 1;
              mbsblockptr->mbs_blocksize = BlockSize;
          } else {
              if (XferMode == 1) {flags = 0;} else {flags = 2;}
              blockptr->hdr_flags = htons(flags);
              blockptr->hdr_stream = htons((short)stream);
              blockptr->hdr_endian = 1;
              blockptr->hdr_spare = 0;
          }
          blockptr->hdr_sequence = 0;
          blockptr->hdr_blocklength = htonl(BlockSize);
          blockptr->hdr_datalength = -1;
          blockptr->hdr_offset = 0;
          blockptr->hdr_id1 = 0;
          blockptr->hdr_id2 = 0;

          len = sizeof(Restart);
       } else {
           if (XferMode == 2 || XferMode == 4) {
              if (stream == 0) return OK;
              blockptr = (HEADER*) data;     /*   XferMode=2 or 4  -  no protocol header   */
           } else {

                  blockptr = (HEADER*) (data - sizeof(HEADER));

/*    32 bytes are available BEFORE the supplied pointer for control information   */
/*    mode = 1 requires ACK for each request; mode = 3 does not  */
/*    mode = 6 (XDAQ/LNL) also uses a 32 byte header */

              if (XferMode == 6) {
                  xdaqblockptr = (I2O_SEND_TO_BU_MESSAGE_FRAME*) (data - sizeof(I2O_SEND_TO_BU_MESSAGE_FRAME));
                  xdaqblockptr->MessageSize = (BlockSize >> 2);
                  xdaqblockptr->MsgFlags = 0;
                  xdaqblockptr->VersionOffset = 0;
                  xdaqblockptr->TargetAddress = MERGE_TID;
                  xdaqblockptr->InitiatorAddress = VXI_TID;
                  xdaqblockptr->Function = I2O_PRIVATE_MESSAGE;
                  xdaqblockptr->XFunctionCode = I2O_BU_GET;
                  xdaqblockptr->OrganizationID = XDAQ_ORGANIZATION_ID;
                  xdaqblockptr->buff_seqn = transferSequence[stream];

/*    now convert the header to little endian - this is 32 / 4 => 8 words    */

                  ptr = (unsigned char *) xdaqblockptr;
                  for (i = 1; i <= 8; i++)
                  {
                      b1 = *ptr++; b2 = *ptr++; b3 = *ptr++; b4 = *ptr++;
                      ptr -=4;
                      *ptr++ = b4; *ptr++ = b3; *ptr++ = b2; *ptr++ = b1;
                  }
              } else {

                  if (XferMode == 1) {flags = 0;} else {flags = 2;}
                  blockptr->hdr_flags = htons(flags);

                  blockptr->hdr_stream = htons((short)stream);
                  blockptr->hdr_endian = 1;
                  blockptr->hdr_spare = htons(0);
                  blockptr->hdr_sequence = htonl(transferSequence[stream]);
                  blockptr->hdr_blocklength = htonl(BlockSize);
                  blockptr->hdr_datalength = htonl(length);
                  blockptr->hdr_offset = htonl(0);
                  blockptr->hdr_id1 = htonl(HDR_ID1);
                  blockptr->hdr_id2 = htonl(HDR_ID2);
              }
           }
           len = BlockSize;
        }

#ifdef USEPTHREAD

      if (length > 0 &&  XferMode != 1) {

        (void) transferTxWait();      //    ensure there is no pending request

//    a suitable request to be passed to the send thread

        if (verbose) printf("transferTxData sending\n");

        do_send_thread(tsSocket,(char *)blockptr,len);

        if (!OverlapMode) transferTxWait();

      } else {

#endif

//    threads are not in use - or an acknowledgement is required 

        retval = do_send(tsSocket,(char *)blockptr,len);
        if (retval == SOCKET_ERROR) {
            ReportError();
            sprintf(MsgBuf, "send() failed: ");
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            transferClose();
            return  ERROR;
        }

#ifdef USEPTHREAD
      }
#endif

        if (verbose) printf("Sent %d bytes to server\n",len);

        if (XferMode == 2 || XferMode == 4) return OK;

        if (stream > 0) {transferSequence[stream]++;}

        if (XferMode == 6) return OK;

        if (flags == 2) {return OK;}   /* ACK is suppressed */

/*    wait for acknowledgement  */

        retval = do_recv(tsSocket,Ack,sizeof (Ack),0 );
        if (retval == SOCKET_ERROR) {
            ReportError();
            sprintf(MsgBuf, "recv() failed: ");
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            transferClose();
            return ERROR;
        }
        if (retval == 0) {
            sprintf(MsgBuf, "Server closed connection\n");
            fprintf (stderr, "%s\n", MsgBuf);
            msgDefineMessage (0, MSG_C_EGLK, MSG_S_ERROR, NULL, MsgBuf);
            transferClose();
            return ERROR;
        }
        if (verbose) printf("Received %d bytes from server\n",retval);

/*    process the acknowledgement   */

        ackptr = (ACK*) &Ack[0];
        if (ntohs(ackptr->acq_code) != 0) return ntohs(ackptr->acq_code);

    return OK;
}
/*******************************************************************************/
